import torch
from torch import Tensor
import torch.nn.functional as F
from torch.fft import rfft, irfft


def _custom_fft_conv1d(input: Tensor, weight: Tensor,
                       stride: int = 1) -> Tensor:
    weight = weight.view(-1, weight.shape[-1])
    output_size = (input.shape[-1] - 1 *
                   (weight.shape[-1] - 1) - 1) // stride + 1

    s = max(input.shape[-1], weight.shape[-1])
    if stride % 2:
        factor = stride * 2
    else:
        factor = stride

    offset = s % factor
    if offset:
        s += factor - offset

    X = rfft(input, n=s)
    W = rfft(weight, n=s)
    # W.imag.mul_(-1)
    torch.view_as_real(W)[..., 1].mul_(-1)
    Y = X * W

    # handle stride
    if stride > 1:
        n_fft = s
        new_n_fft = n_fft // stride
        step_size = new_n_fft // 2
        strided_Y_size = step_size + 1

        tmpY = torch.view_as_real(Y)
        Y_real, Y_imag = tmpY[..., 0], tmpY[..., 1]

        unfolded_Y_real = Y_real.unfold(-1, strided_Y_size, step_size)
        unfolded_Y_imag = Y_imag[...,
                                 1:].unfold(-1, strided_Y_size - 2, step_size)
        Y_pos_real, Y_pos_imag = unfolded_Y_real[..., ::2,
                                                 :].sum(-2), unfolded_Y_imag[..., ::2, :].sum(-2)
        Y_neg_real, Y_neg_imag = unfolded_Y_real[..., 1::2, :].sum(
            -2).flip(-1), unfolded_Y_imag[..., 1::2, :].sum(-2).flip(-1)

        Y_real = Y_pos_real.add_(Y_neg_real)
        Y_imag = Y_pos_imag.add_(Y_neg_imag, alpha=-1)
        Y_imag = F.pad(Y_imag, [1, 1])

        Y = torch.view_as_complex(
            torch.stack((Y_real, Y_imag), -1)).div_(stride)

    output = irfft(Y)

    # Remove extra padded values
    output = output[..., :output_size].contiguous()
    return output



import torch


def sinc_kernel(num_zeros, precision, roll_off=0.945, dtype=torch.float64):
    """
    Sinc interpolation kernel.

    Args:
        num_zeros (int): number of zero crossing to keep in the sinc filter
        precision (int): number of filter coefficients to retain for each zero-crossing
        roll_off (float): the roll-off frequency (as a fraction of nyquist). Default: ``0.945``
        dtype (torch.dtype): Default: ``torch.float64``
    """
    t = torch.arange(-num_zeros * precision, num_zeros *
                     precision + 1, dtype=dtype) / precision * roll_off
    return torch.sinc(t) * roll_off


if __name__ == '__main__':
    print(sinc_kernel(16, 2))



import torch
from torch import nn
import torch.nn.functional as F
from typing import Callable, Tuple
from .sinc import sinc_kernel
from .fftconv import _custom_fft_conv1d


def _manual_reflect_pad(x: torch.Tensor, padding: Tuple[int, int]):
    flipped_x = x.flip(-1)
    tensors = [x]
    left_padding, right_padding = padding
    step = False
    while left_padding > 0:
        if step:
            tensors.insert(0, x[..., -left_padding-1:-1])
        else:
            tensors.insert(0, flipped_x[..., -left_padding-1:-1])
        left_padding -= x.shape[-1] - 1
        step = not step

    step = False
    while right_padding > 0:
        if step:
            tensors.append(x[..., 1:right_padding + 1])
        else:
            tensors.append(flipped_x[..., 1:right_padding + 1])
        right_padding -= x.shape[-1] - 1
        step = not step
    x = torch.cat(tensors, -1)
    return x


def _pad_to_block_2(x: torch.Tensor, block_size: int, padding: int):
    offset = x.shape[-1] % block_size
    if offset:
        offset = block_size - offset
    paddings = (padding, offset + padding)
    if max(paddings) >= x.shape[-1]:
        x = _manual_reflect_pad(x, paddings)
    else:
        x = F.pad(x, paddings, mode='reflect')
    return x.unfold(-1, block_size + 2 * padding, block_size)


class Upsample(nn.Module):
    r"""Upsampling by an integer amount.

    Args:
        q (int): upsample factor
        num_zeros (int): number of zero crossing to keep in the sinc filter. Default: ``16``
        window_func (Callable): window function. Default: :meth:`hann_window <torch.hann_window>`
        **kwargs: arguments passed through to :meth:`sinc_kernel <torch.sinc_kernel>`

    Shape:
        - Input: `[*, T]`
        - Output: `[*, T * q]`

    Examples::

        >>> import torch
        >>> upsampler = Upsample(3)
        >>> x = torch.randn(500)
        >>> print(len(upsampler(x)))
        1500

    """
    __constants__ = ['BLOCK_RATIO']
    BLOCK_RATIO: int = 5

    def __init__(self,
                 q: int = 2,
                 num_zeros: int = 16,
                 window_func: Callable[[int],
                                       torch.Tensor] = torch.hann_window,
                 **kwargs):
        assert q > 1, "q should be greater than 1"
        super().__init__()
        kernel = sinc_kernel(num_zeros, q, **kwargs)
        N = kernel.numel()
        kernel *= window_func(N, dtype=torch.float64)
        kernel = F.pad(kernel.float(), [q - 1, 0]).view(-1,
                                                        q).t().reshape(q, 1, -1).flip(0)
        self.stride = q
        self.padding = N // 2 // q
        self.register_buffer('kernel', kernel)

    def forward(self, x: torch.Tensor):
        shape = x.shape
        x = x.view(-1, 1, shape[-1])

        block_length = self.kernel.shape[-1] * self.BLOCK_RATIO
        if shape[-1] < block_length:
            x = _manual_reflect_pad(x, (self.padding, self.padding))
            y = _custom_fft_conv1d(x, self.kernel)
        else:
            q = self.kernel.shape[0]
            x = _pad_to_block_2(x, block_length, self.padding)
            num_blocks = x.shape[-2]
            x = x.reshape(-1, 1, x.shape[-1])
            y = _custom_fft_conv1d(x, self.kernel)
            y = y.view(-1, num_blocks, q, block_length).transpose(1,
                                                                  2).reshape(-1, q, num_blocks * block_length)[..., :shape[-1]]

        y = y.transpose(1, 2).contiguous()
        return y.view(shape[:-1] + (-1,))



import torch
import math
import torch.nn.functional as F

from .fftconv import _custom_fft_conv1d


def _pad_to_block(x, block_size, padding):
    offset = x.shape[-1] % block_size
    if offset:
        offset = block_size - offset
    x = F.pad(x, [padding[0], offset + padding[1]], mode='replicate')
    return x.unfold(-1, block_size + sum(padding), block_size)


class Resample(torch.nn.Module):
    """
    Resampling from the sample rate `old_sr` to `new_sr`.
    """

    def __init__(self, old_sr: int, new_sr: int, zeros: int = 24, rolloff: float = 0.945):
        """
        Args:
            old_sr (int): sample rate of the input signal x.
            new_sr (int): sample rate of the output.
            zeros (int): number of zero crossing to keep in the sinc filter.
            rolloff (float): use a lowpass filter that is `rolloff * new_sr / 2`,
                to ensure sufficient margin due to the imperfection of the FIR filter used.
                Lowering this value will reduce anti-aliasing, but will reduce some of the
                highest frequencies.

        Shape:

            - Input: `[*, T]`
            - Output: `[*, T']` with `T' = int(new_sr * T / old_sr)


        .. caution::
            After dividing `old_sr` and `new_sr` by their GCD, both should be small
            for this implementation to be fast.

        >>> import torch
        >>> resample = ResampleFrac(4, 5)
        >>> x = torch.randn(1000)
        >>> print(len(resample(x)))
        1250
        """
        super().__init__()
        if not isinstance(old_sr, int) or not isinstance(new_sr, int):
            raise ValueError("old_sr and new_sr should be integers")
        gcd = math.gcd(old_sr, new_sr)
        self.old_sr = old_sr // gcd
        self.new_sr = new_sr // gcd
        self.zeros = zeros
        self.rolloff = rolloff

        self._init_kernels()

    def _init_kernels(self):
        if self.old_sr == self.new_sr:
            return

        kernels = []
        sr = min(self.new_sr, self.old_sr)
        # rolloff will perform antialiasing filtering by removing the highest frequencies.
        # At first I thought I only needed this when downsampling, but when upsampling
        # you will get edge artifacts without this, the edge is equivalent to zero padding,
        # which will add high freq artifacts.
        sr *= self.rolloff

        # The key idea of the algorithm is that x(t) can be exactly reconstructed from x[i] (tensor)
        # using the sinc interpolation formula:
        #   x(t) = sum_i x[i] sinc(pi * old_sr * (i / old_sr - t))
        # We can then sample the function x(t) with a different sample rate:
        #    y[j] = x(j / new_sr)
        # or,
        #    y[j] = sum_i x[i] sinc(pi * old_sr * (i / old_sr - j / new_sr))

        # We see here that y[j] is the convolution of x[i] with a specific filter, for which
        # we take an FIR approximation, stopping when we see at least `zeros` zeros crossing.
        # But y[j+1] is going to have a different set of weights and so on, until y[j + new_sr].
        # Indeed:
        # y[j + new_sr] = sum_i x[i] sinc(pi * old_sr * ((i / old_sr - (j + new_sr) / new_sr))
        #               = sum_i x[i] sinc(pi * old_sr * ((i - old_sr) / old_sr - j / new_sr))
        #               = sum_i x[i + old_sr] sinc(pi * old_sr * (i / old_sr - j / new_sr))
        # so y[j+new_sr] uses the same filter as y[j], but on a shifted version of x by `old_sr`.
        # This will explain the F.conv1d after, with a stride of old_sr.
        self._width = math.ceil(self.zeros * self.old_sr / sr)
        # If old_sr is still big after GCD reduction, most filters will be very unbalanced, i.e.,
        # they will have a lot of almost zero values to the left or to the right...
        # There is probably a way to evaluate those filters more efficiently, but this is kept for
        # future work.
        idx = torch.arange(-self._width, self._width + self.old_sr).float()
        for i in range(self.new_sr):
            t = (-i/self.new_sr + idx/self.old_sr) * sr
            t = t.clamp_(-self.zeros, self.zeros)
            t *= math.pi
            window = torch.cos(t/self.zeros/2)**2
            kernel = torch.sinc(t / math.pi) * window
            # Renormalize kernel to ensure a constant signal is preserved.
            kernel.div_(kernel.sum())
            kernels.append(kernel)

        self.register_buffer("kernel", torch.stack(
            kernels).view(self.new_sr, 1, -1))

    def forward(self, x: torch.Tensor):
        if self.old_sr == self.new_sr:
            return x
        shape = x.shape
        length = x.shape[-1]
        x = x.reshape(-1, length)

        block_length = self.kernel.shape[-1] * self.old_sr * 2
        if length < block_length:
            x = F.pad(x[:, None], (self._width, self._width +
                                   self.old_sr), mode='replicate')
            ys = _custom_fft_conv1d(x, self.kernel, stride=self.old_sr)
        else:
            x = _pad_to_block(x[:, None], block_length, (self._width, self._width +
                                                         self.old_sr))
            num_blocks = x.shape[-2]
            x = x.reshape(-1, 1, x.shape[-1])
            ys = _custom_fft_conv1d(x, self.kernel, stride=self.old_sr)
            ys = ys.view(-1, num_blocks, self.new_sr, ys.shape[-1]).transpose(
                1, 2).reshape(-1, self.new_sr, num_blocks * ys.shape[-1])
        y = ys.transpose(1, 2).reshape(list(shape[:-1]) + [-1])
        return y[..., :int(self.new_sr * length / self.old_sr)]



import torch
from torch import nn
import torch.nn.functional as F
from typing import Callable

from .sinc import sinc_kernel
from .upsample import _pad_to_block_2, _manual_reflect_pad
from .fftconv import _custom_fft_conv1d


class Decimate(nn.Module):
    r"""Downsampling by an integer amount.

    Args:
        q (int): downsample factor
        num_zeros (int): number of zero crossing to keep in the sinc filter. Default: ``16``
        window_func (Callable): window function. Default: :meth:`hann_window <torch.hann_window>`
        **kwargs: arguments passed through to :meth:`sinc_kernel <torch.sinc_kernel>`

    Shape:
        - Input: `[*, T]`
        - Output: `[*, T / q]`

    Examples::

        >>> import torch
        >>> decimater = Decimate(4)
        >>> x = torch.randn(1000)
        >>> print(len(decimater(x)))
        250

    """
    __constants__ = ['BLOCK_RATIO']
    BLOCK_RATIO: int = 4

    def __init__(self,
                 q: int = 2,
                 num_zeros: int = 16,
                 window_func: Callable[[int],
                                       torch.Tensor] = torch.hann_window,
                 **kwargs):
        assert q > 1, "q should be greater than 1"
        super().__init__()
        kernel = sinc_kernel(num_zeros, q, **kwargs)
        N = kernel.numel()
        kernel *= window_func(N, dtype=torch.float64)
        kernel /= q

        self.stride = q
        self.padding = N // 2
        self.register_buffer('kernel', kernel.view(1, 1, -1).float())

    def forward(self, x: torch.Tensor):
        shape = x.shape
        x = x.view(-1, 1, shape[-1])

        block_length = self.kernel.shape[-1] * self.stride * self.BLOCK_RATIO
        out_size = shape[-1] // self.stride
        if shape[-1] < block_length:
            x = _manual_reflect_pad(x, (self.padding, self.padding))
            y = _custom_fft_conv1d(x, self.kernel, stride=self.stride)
        else:
            x = _pad_to_block_2(x, block_length, self.padding)
            num_blocks = x.shape[-2]
            x = x.reshape(-1, 1, x.shape[-1])
            y = _custom_fft_conv1d(x, self.kernel, stride=self.stride)
            y = y.view(-1, num_blocks * y.shape[-1])[..., :out_size]

        return y.view(shape[:-1] + (-1,))




from kazane import Upsample
import torch
import pytest

from .test_decimate import make_sweep, make_tone

device = 'cpu'
dtype = torch.float64


@pytest.mark.parametrize('q', [2, 3, 5, 7])
@pytest.mark.parametrize('zeros,rms', [
    (64, 1e-4),
    (32, 1e-4),
    (16, 4e-4)
])
def test_quality_sine(q, zeros, rms):
    FREQ = 512.0
    DURATION = 2
    sr = 22050
    sr_new = sr * q
    up = torch.jit.script(Upsample(q, zeros,
                                   roll_off=0.945)).to(dtype)

    x = make_tone(FREQ, sr, DURATION)
    y = make_tone(FREQ, sr_new, DURATION)
    y_pred = up(x)

    idx = slice(sr_new // 2, - sr_new // 2)

    err = torch.mean(torch.abs(y[idx] - y_pred[idx])).item()
    assert err <= rms, '{:g} > {:g}'.format(err, rms)


@pytest.mark.parametrize('q', [2, 3])
@pytest.mark.parametrize('zeros,rms', [
    (64, 5e-2),
    (32, 5e-2),
    (16, 5e-2)
])
def test_quality_sweep(q, zeros, rms):
    FREQ = 4096
    DURATION = 5
    sr = 22050
    sr_new = sr * q
    up = torch.jit.script(Upsample(q, zeros,
                                   roll_off=0.945)).to(dtype)

    x = make_sweep(FREQ, sr, DURATION)
    y = make_sweep(FREQ, sr_new, DURATION)
    y_pred = up(x)

    idx = slice(sr_new // 2, - sr_new // 2)

    err = torch.mean(torch.abs(y[idx] - y_pred[idx])).item()
    assert err <= rms, '{:g} > {:g}'.format(err, rms)


def test_short_input():
    up = Upsample(2, 32)
    x = torch.randn(8)
    y = up(x)



from kazane import Decimate
import torch
import pytest
from math import pi, log2

device = 'cpu'
dtype = torch.float64


def make_tone(freq, sr, duration):
    return torch.sin(2 * pi * freq / sr * torch.arange(int(sr * duration), dtype=dtype))


def make_sweep(freq, sr, duration):
    return torch.sin(torch.cumsum(2 * pi * torch.logspace(log2(2.0 / sr),
                                                          log2(
                                                              float(freq) / sr),
                                                          steps=int(duration*sr), base=2.0, dtype=dtype), dim=-1))


@pytest.mark.parametrize('q', [2, 3, 5, 7])
@pytest.mark.parametrize('zeros,rms', [
    (64, 2e-4),
    (32, 2e-4),
    (16, 4e-4)
])
def test_quality_sine(q, zeros, rms):
    FREQ = 512.0
    DURATION = 2
    sr = 44100
    sr_new = sr // q
    dec = torch.jit.script(Decimate(q, zeros, roll_off=0.945)).to(dtype)

    x = make_tone(FREQ, sr, DURATION)
    y = make_tone(FREQ, sr_new, DURATION)
    y_pred = dec(x)

    idx = slice(sr_new // 2, - sr_new // 2)

    err = torch.mean(torch.abs(y[idx] - y_pred[idx])).item()
    assert err <= rms, '{:g} > {:g}'.format(err, rms)


@pytest.mark.parametrize('q', [2, 3])
@pytest.mark.parametrize('zeros,rms', [
    (64, 8e-2),
    (32, 8e-2),
    (16, 8e-2)
])
def test_quality_sweep(q, zeros, rms):
    FREQ = 4096
    DURATION = 5
    sr = 44100
    sr_new = sr // q
    dec = torch.jit.script(Decimate(q, zeros, roll_off=0.945)).to(dtype)

    x = make_sweep(FREQ, sr, DURATION)
    y = make_sweep(FREQ, sr_new, DURATION)
    y_pred = dec(x)

    idx = slice(sr_new // 2, - sr_new // 2)

    err = torch.mean(torch.abs(y[idx] - y_pred[idx])).item()
    assert err <= rms, '{:g} > {:g}'.format(err, rms)


def test_short_input():
    decimater = Decimate(4)
    x = torch.randn(8)
    y = decimater(x)



import setuptools

NAME = "kazane"
VERSION = '1.0.3'
AUTHOR = 'Chin-Yun Yu'
EMAIL = 'lolimaster.cs03@nctu.edu.tw'


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    author_email=EMAIL,
    description="",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/yoyololicon/kazane",
    packages=setuptools.find_packages(),
    install_requires=['torch'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from kazane.upsample import _pad_to_block_2
import pytest
import torch


@pytest.mark.parametrize('x_shape', [
    (1, 1, 16),
    (8, 4, 32)
])
@pytest.mark.parametrize('block_size', [8, 16, 32, 64])
@pytest.mark.parametrize('padding', [4, 16, 64])
def test_pad_to_block(x_shape, block_size, padding):
    x = torch.randn(*x_shape)
    _pad_to_block_2(x, block_size, padding)
